package com.perforce.p4simulink;

/**
 * Copyright (C) 2014 Perforce Software. All rights reserved.
 *
 * Please see LICENSE.txt in top-level folder of this distribution.
 */

import java.io.File;
import java.io.IOException;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.List;
import java.util.Properties;

import javax.swing.JOptionPane;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import com.mathworks.cmlink.api.ApplicationInteractor;
import com.mathworks.cmlink.api.ConfigurationManagementException;
import com.mathworks.cmlink.api.InteractorSupportedFeature;
import com.mathworks.cmlink.api.Terminator;
import com.mathworks.cmlink.api.customization.CoreAction;
import com.mathworks.cmlink.api.customization.CustomizationWidgetFactory;
import com.mathworks.cmlink.api.version.r14a.CMInteractor;
import com.perforce.p4java.PropertyDefs;
import com.perforce.p4java.client.IClient;
import com.perforce.p4java.core.file.FileSpecBuilder;
import com.perforce.p4java.core.file.IFileSpec;
import com.perforce.p4java.exception.P4JavaException;
import com.perforce.p4java.impl.generic.client.ClientView;
import com.perforce.p4java.impl.mapbased.client.Client;
import com.perforce.p4java.impl.mapbased.rpc.RpcPropertyDefs;
import com.perforce.p4java.server.IOptionsServer;
import com.perforce.p4java.server.ServerFactory;
import com.perforce.p4java.server.ServerStatus;
import com.perforce.p4simulink.connection.Config;
import com.perforce.p4simulink.ui.P4ConnectionPane;
import com.perforce.p4simulink.ui.P4PasswordPane;
import com.perforce.p4simulink.util.Logging;

public class P4Interactor implements CMInteractor {

	private static final int OK = 0;
	private static final int CANCEL = 1;
	private static final int ERROR = 2;

	private static final Properties properties;
	protected static final Logger log = LogManager.getLogger(P4Interactor.class
			.getName());

	// list of supported features
	private static final Collection<InteractorSupportedFeature> supportedFeatures = EnumSet
			.of(InteractorSupportedFeature.CONNECTION);

	// static properties for all perforce connections
	static {
		properties = System.getProperties();
		properties.put(PropertyDefs.PROG_NAME_KEY, "P4Simulink");
		properties.put(PropertyDefs.PROG_VERSION_KEY, "0.2.991735");

		// Set up socket pooling to use a single socket
		properties.put(RpcPropertyDefs.RPC_SOCKET_POOL_SIZE_NICK, "1");
		// disable timeout for slow servers / large db lock times
		properties.put(RpcPropertyDefs.RPC_SOCKET_SO_TIMEOUT_NICK, "0");
	}
	protected IOptionsServer server = null;
	protected IClient client = null;
	protected Config config;
	protected Terminator terminator;
	protected ApplicationInteractor applicationInteractor;
	protected File sandboxRoot;
	protected String depotPath;

	public P4Interactor(ApplicationInteractor applicationInteractor,
			Config config) throws P4CMException {
		// load config
		this.config = config;
		if (config != null && config.exists()) {
			this.config.load();
		}

		// this is a handle we can use to open new dialog/popup windows if
		// needed
		this.applicationInteractor = applicationInteractor;

		terminator = applicationInteractor.getTerminator();
	}

	public P4Interactor setDepotPath(String path) {
		depotPath = path;
		return this;
	}

	public String getDepotPath() {
		return depotPath;
	}

	public void setTerminator(Terminator terminator) {
		this.terminator = terminator;
	}

	public File getSandboxRoot() {
		return sandboxRoot;
	}

	public void setSandboxRoot(File sandboxRoot) {
		this.sandboxRoot = sandboxRoot;
	}

	@Override
	public void buildCustomActions(CustomizationWidgetFactory widgetFactory) {
		// create a button which will allow the user to create a
		// sandbox-specific .p4config
		widgetFactory.createActionWidget("Manage .p4config", null,
				new CoreAction() {
					@Override
					public void execute()
							throws ConfigurationManagementException {
						showConnectDialogue("");
					}

					@Override
					public String getDescription() {
						return "Manage .p4config";
					}

					@Override
					public boolean canCancel() {
						return true;
					}
				});
	}

	@Override
	public boolean isFeatureSupported(
			InteractorSupportedFeature interactorSupportedFeature) {
		return supportedFeatures.contains(interactorSupportedFeature);
	}

	/**
	 * Checks if the integration is ready. As the integration needs to be
	 * connected to a Perforce Server, we check if a IOptionsServer object
	 * exists for the session and is connected. If it does, then we're ready;
	 * otherwise, we still need to connect.
	 *
	 * @return is the connection to the Perforce Server ready or not.
	 */
	@Override
	public boolean isReady() {
		return server != null && server.isConnected()
				&& server.getStatus() == ServerStatus.READY;
	}

	/**
	 * Disconnect from the Perforce Server and close the Perforce session.
	 *
	 * @throws ConfigurationManagementException
	 */
	@Override
	public void disconnect() throws ConfigurationManagementException {
		try {
			if (isReady()) {
				server.disconnect();
			}
		} catch (Exception e) {
			ConfigurationManagementException ex = new ConfigurationManagementException(
					e);
			log.error(ex);
			throw ex;
		}
	}

	public IClient getClient(String clientName)
			throws ConfigurationManagementException {
		return getClient(clientName, false);
	}

	/**
	 * @param clientName
	 *            specified client name
	 * @return IClient client
	 */
	public IClient getClient(String clientName, boolean createIfNeeded)
			throws ConfigurationManagementException {
		// if we're already using the correct client, return it
		if (client != null && client.getName().equals(clientName)) {
			return client;
		}

		try {
			if (server == null) {
				throw new P4CMException("Unable to connect to server.");
			}

			if (clientName == null || clientName.isEmpty()) {
				throw new P4CMException(
						"No client specified. Please supply a valid client.");
			}

			IClient newClient = server.getClient(clientName);

			// specified client doesn't exist, so pop an error
			if (newClient == null) {
				if (!createIfNeeded) {
					throw new P4CMException("Client '" + clientName
							+ "' not found.");
				}

				// client is still null and we're set to create if needed
				newClient = new Client();
				newClient.setName(clientName);
				newClient.setDescription("Created by p4simulink (0.2.991735)");
				newClient.setHostName("");
				newClient.setServer(server);

				// client root
				newClient.setRoot(sandboxRoot.getAbsolutePath());

				// set a default view
				ClientView view = new ClientView();
				ClientView.ClientViewMapping mapping = new ClientView.ClientViewMapping(
						0, depotPath, "//" + clientName + "/...");
				view.addEntry(mapping);
				newClient.setClientView(view);

				server.createClient(newClient);
				server.setCurrentClient(newClient);
				log.debug("Created client '" + clientName + "'");
			}

			// ensure the client root exists
			File clientRoot = new File(newClient.getRoot());
			if (!clientRoot.exists() || !clientRoot.isDirectory()) {
				throw new P4CMException(
						"Specified client root directory does not exist for client '"
								+ clientName + "'");
			}

			// ensure the client root contains our sandbox
			File current = new File(sandboxRoot.getAbsolutePath());
			boolean isContained = false;
			while (current != null) {
				if (current.equals(clientRoot)) {
					isContained = true;
					break;
				}
				current = current.getParentFile();
			}

			if (!isContained) {
				throw new P4CMException(
						"Specified client root directory does not contain the specified sandbox directory.");
			}

			// set our client to be the current one
			client = newClient;
			server.updateClient(client, false);
			server.setCurrentClient(client);
			client.setServer(server);
			return client;
		} catch (P4JavaException e) {
			Logging.logException(log, e, true);
		}

		return client;
	}

	public IClient getClient() throws ConfigurationManagementException {
		try {
			return getClient(config.getClient());
		} catch (ConfigurationManagementException e) {
			Logging.logException(log, e, true);
		}

		return null;
	}

	public IOptionsServer getServer() {
		return server;
	}

	/**
	 * Converts a collection of files into a list of IFileSpec objects. Any
	 * directories contained in the input list are expanded into a list of files
	 * contained within them (if any).
	 *
	 * @param files
	 *            Collection of files for which we want to get file specs. Note
	 *            that directories are expanded into the list of files contain
	 *            in them (if any).
	 * @return List of IFileSpec objects for the input list.
	 */
	public List<IFileSpec> toSpec(Collection<File> files) {
		List<IFileSpec> specs = new ArrayList<>();
		try {
			HashSet<String> filePaths = new HashSet<>();
			for (File file : files) {
				// adding a directory means we should add the files under it
				if (file.isDirectory()) {
					filePaths.add(file.getCanonicalPath() + File.separator
							+ "...");
				} else {

					String filename = file.getCanonicalPath();
					if (filename.contains("#") || filename.contains("@")
							|| filename.contains("%")) {
						filename = filename.replace("%", "%25");
						filename = filename.replace("@", "%40").replace("#",
								"%23");
					}

					// regular file
					filePaths.add(filename);
				}
			}

			if (!filePaths.isEmpty()) {
				specs = FileSpecBuilder.makeFileSpecList(filePaths
						.toArray(new String[filePaths.size()]));
			}
		} catch (IOException e) {
			log.error("IO Exception building list of file specs: "
					+ e.getLocalizedMessage());
		}
		return specs;
	}

	public List<IFileSpec> toSpec(File file) {
		return toSpec(Collections.singletonList(file));
	}

	/**
	 * Get the address from our P4Config object and attempt to connect to that
	 * endpoint. If there are any problems connecting to the server, catch the
	 * exception and rethrow as a P4CMException.
	 *
	 * @throws ConfigurationManagementException
	 */
	@Override
	public void connect() throws ConfigurationManagementException {
		log.debug("connect()");

		// not enough information to connect to Perforce
		if (config == null || config.getP4JavaURI().isEmpty()
				|| config.getUser().isEmpty() || config.getClient().isEmpty()) {
			return;
		}

		// connect to server
		try {
			// try connecting to the server with the info we have
			server = ServerFactory.getOptionsServer(config.getP4JavaURI(),
					properties);
			server.connect();
			server.setUserName(config.getUser());
		} catch (URISyntaxException e) {
			// connection failed
			Logging.logException(log, new ConfigurationManagementException(e),
					false);
		} catch (P4JavaException e) {
			// connection failed
			Logging.logException(log, e, false);
		}
		/*
		 * // now get our global client if (server != null &&
		 * server.isConnected() && server.getStatus() == ServerStatus.READY) {
		 * client = getClient(config.getClient()); }
		 */
	}

	private int showConnectDialogue(String results) throws P4CMException {
		P4ConnectionPane pane = new P4ConnectionPane(config);
		if (results != null && !results.isEmpty()) {
			pane.setResults(results);
		}
		int option = JOptionPane.showConfirmDialog(
				applicationInteractor.getParentFrame(), pane.getFields(),
				"Open Connection", JOptionPane.OK_CANCEL_OPTION);

		if (option == JOptionPane.CANCEL_OPTION) {
			// user hit cancel
			return CANCEL;
		}

		boolean isValid = pane.getServer() != null
				&& !pane.getServer().isEmpty() && pane.getUsername() != null
				&& !pane.getUsername().isEmpty() && pane.getWorkspace() != null
				&& !pane.getWorkspace().isEmpty();

		// something was invalid
		if (!isValid) {
			return ERROR;
		}

		// we're good, so update the config values and save
		config.setPort(pane.getServer());
		config.setClient(pane.getWorkspace());
		config.setUser(pane.getUsername());

		try {
			config.save();
		} catch (P4CMException e) {
			Logging.logException(log, new P4CMException(
					"Couldn't save config: " + e.getLocalizedMessage()), true);
		}

		return OK;
	}

	public int showPasswordDialogue(String result) {
		P4PasswordPane pane = new P4PasswordPane();

		if (result != null && !result.isEmpty()) {
			pane.setResults(result);
		}

		int option = JOptionPane.showConfirmDialog(
				applicationInteractor.getParentFrame(), pane.getFields(),
				"Enter Password", JOptionPane.OK_CANCEL_OPTION);

		if (option == JOptionPane.CANCEL_OPTION) {
			return CANCEL;
		}

		// update the config with the new values, if the user clicked ok
		config.setPassword(pane.getPassword());
		return OK;
	}

	@Override
	public String getSystemName() {
		return "Perforce SCM Integration for Simulink [p4simulink (0.2.991735)]";
	}

	@Override
	public String getShortSystemName() {
		return "p4simulink (0.2.991735)";
	}

}
